/*
 Copyright (c) 2001-2002-2003 sync4j project
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.
 
 3. The name "sync4j" must not be used to endorse or promote products
    derived from this software without prior written permission.
 
 4. Products derived from this software may not be called "sync4j", nor
    may "sync4j" appear in their name, without prior written permission.
 
 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
 
     "This product includes software developed by the
      sync4j project."
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE SYNC4J AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE. 
 */

package sync4j.server.store;

import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Set;
import java.util.Iterator;

import java.sql.*;

import sync4j.framework.tools.DBTools;
import sync4j.framework.server.store.*;

import sync4j.framework.security.Sync4jPrincipal;
import sync4j.server.engine.Sync4jSource;

import org.apache.commons.lang.StringUtils;

/**
 * This is the store for information regarding the Sync4j engine such as users,
 * principals, ecc. Currently it persistes the following classes:
 *
 * <ul>
 * <li>sync4j.framework.security.Sync4jPrincipal</li>
 * </ul>
 *
 *
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: EnginePersistentStore.java,v 1.6 2003/10/15 13:20:02 luigiafassina Exp $
 *
 */
public class EnginePersistentStore 
extends BasePersistentStore
implements PersistentStore, java.io.Serializable {
    
    // --------------------------------------------------------------- Constants

    public static final int SQL_INSERT_PRINCIPAL      = 0;
    public static final int SQL_GET_PRINCIPAL         = 1;
    public static final int SQL_SELECT_PRINCIPAL      = 2;
    public static final int SQL_UPDATE_PRINCIPAL      = 3;
    public static final int SQL_SELECT_ALL_PRINCIPALS = 4;
    public static final int SQL_INSERT_SOURCE         = 5;
    public static final int SQL_GET_SOURCE            = 6;
    public static final int SQL_UPDATE_SOURCE         = 7;
    public static final int SQL_SELECT_ALL_SOURCES    = 8;
    // -------------------------------------------------------------- Properties    
    
    private String[] sql = null;
    
    public void setSql(String[] sql) {
        this.sql = sql;
    }
    
    public String[] getSql() {
        return this.sql;
    }

    // ------------------------------------------------------------ Constructors
    // ---------------------------------------------------------- Public methods
    
    public boolean store(Object o)
    throws PersistentStoreException {
        if (o instanceof Sync4jPrincipal) {
            storePrincipal((Sync4jPrincipal) o);
            return true;
        } else if (o instanceof Sync4jSource) {
            storeSource((Sync4jSource) o);
            return true;
        }
        
        return false;
    }
    
    public boolean read(Object o)
    throws PersistentStoreException {
        if (o instanceof Sync4jPrincipal) {
            readPrincipal((Sync4jPrincipal) o);
            return true;
        } else if (o instanceof Sync4jSource) {
            readSource((Sync4jSource) o);
            return true;
        }
        
        return false;
    }
    
    public Object[] read(Class objClass) throws PersistentStoreException {
        if (objClass.equals(Sync4jPrincipal.class)) {
            return readAllPrincipals();
        } else if (objClass.equals(Sync4jSource.class)) {
            return readAllSources();
        }
        
        return null;
    }

    
    // --------------------------------------------------------- Private methods
    
    private void storePrincipal(Sync4jPrincipal p)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        
        try {
            conn = dataSource.getConnection();
                        
            stmt = conn.prepareStatement(sql[SQL_UPDATE_PRINCIPAL]);
            
            stmt.setString(1, p.getUsername());
            stmt.setString(2, p.getDeviceId());
            stmt.setString(3, p.getId()      );
            
            int n = stmt.executeUpdate();
            
            if (n == 0) {
                //
                // The first time!!!
                //
                stmt.close();
                
                stmt = conn.prepareStatement(sql[SQL_INSERT_PRINCIPAL]);
                
                stmt.setString(1, p.getId()      );
                stmt.setString(2, p.getUsername());
                stmt.setString(3, p.getDeviceId());
                
                stmt.executeUpdate();
            }
        } catch (SQLException e) {
            throw new PersistentStoreException("Error storing principal " + p, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }

    /**
     * Read the principal from the data store. If <i>getId()</i> returns null, 
     * it tries to read the principal from username/device, otherwise throws id
     * is used for the lookup.
     *
     * @throws PersistentException in case of error reading the data store
     */
    private void readPrincipal(Sync4jPrincipal p)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
                
        try {
            conn = dataSource.getConnection();
            
            if (p.getId() == null) {
                
                stmt = conn.prepareStatement(sql[SQL_SELECT_PRINCIPAL]);   
                
                stmt.setString(1, p.getUsername());
                stmt.setString(2, p.getDeviceId());
            } else {
                
                stmt = conn.prepareStatement(sql[SQL_GET_PRINCIPAL]);
                
                stmt.setString(1, p.getId());
            }

            rs = stmt.executeQuery();
            
            if (rs.next() == false) {
                throw new NotFoundException("Principal not found for "
                + p.toString()
                );
            }
            
            p.setId       (rs.getString(1));
            p.setUsername (rs.getString(2));
            p.setDeviceId (rs.getString(3));
            p.setEmail    (rs.getString(4));
            p.setFirstName(rs.getString(5));
            p.setLastName (rs.getString(6));
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading principal " + p, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }
    
    /**
     * Read all principals
     *
     * @return an array with the principals (empty if no objects are found)
     *
     * @throws PersistentException in case of error reading the data store
     */
    private Object[] readAllPrincipals()
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
          
        ArrayList ret = new ArrayList();
        
        try {
            conn = dataSource.getConnection();
            
            stmt = conn.prepareStatement(sql[SQL_SELECT_ALL_PRINCIPALS]);

            rs = stmt.executeQuery();
            
            Sync4jPrincipal p = null;
            while (rs.next()) {
                ret.add(
                    new Sync4jPrincipal(
                        rs.getString(1),
                        rs.getString(2),
                        rs.getString(3)
                    )
                );
            }
            
            return ret.toArray(new Sync4jPrincipal[ret.size()]);
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading principals", e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }
    
    private void storeSource(Sync4jSource s)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        
        try {
            conn = dataSource.getConnection();
                        
            stmt = conn.prepareStatement(sql[SQL_UPDATE_SOURCE]);
            
            stmt.setString(1, s.getConfig());
            stmt.setString(2, s.getUri()   );
                        
            int n = stmt.executeUpdate();
            
            if (n == 0) {
                //
                // The first time!!!
                //
                stmt.close();
                
                stmt = conn.prepareStatement(sql[SQL_INSERT_SOURCE]);
                
                stmt.setString(1, s.getUri()   );
                stmt.setString(2, s.getConfig());
                
                stmt.executeUpdate();
            }
        } catch (SQLException e) {
            throw new PersistentStoreException("Error storing the source " + s, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }

    /**
     * Read the source from the data store. 
     *
     * @param s the source
     *
     * @throws PersistentException in case of error reading the data store
     */
    private void readSource(Sync4jSource s)
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
                
        try {
            conn = dataSource.getConnection();
            
            stmt = conn.prepareStatement(sql[SQL_GET_SOURCE]);
            stmt.setString(1, s.getUri());

            rs = stmt.executeQuery();
            
            if (rs.next() == false) {
                throw new NotFoundException("Source not found for "
                + s.getUri()
                );
            }
            
            s.setName  (rs.getString(1));
            s.setConfig(rs.getString(2));
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading the source " + s, e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }
    
    /**
     * Read all principals
     *
     * @return an array with the principals (empty if no objects are found)
     *
     * @throws PersistentException in case of error reading the data store
     */
    private Object[] readAllSources()
    throws PersistentStoreException {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
          
        ArrayList ret = new ArrayList();
        
        try {
            conn = dataSource.getConnection();
            
            stmt = conn.prepareStatement(sql[SQL_SELECT_ALL_SOURCES]);

            rs = stmt.executeQuery();
            
            Sync4jSource s = null;
            while (rs.next()) {
                ret.add(
                    new Sync4jSource(
                        rs.getString(1),
                        rs.getString(2)
                    )
                );
            }
            
            return ret.toArray(new Sync4jSource[ret.size()]);
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading sources", e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
    }
}
